#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <unistd.h>
#include <ctype.h>
#include <stdlib.h>
#include <dirent.h>
#include <signal.h>
#include <sys/time.h>
#include <utime.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/socket.h>
#include <sys/un.h>
#include "litefs.h"
#include "litefs.p"

static bool debug=false;

static void n_debug (const char *ctl, ...)
{
	if (debug){
		va_list list;
		va_start (list,ctl);
		vfprintf (stderr,ctl,list);
		va_end (list);
	}
}


static void usage()
{
	fprintf (stderr,"litefsd version 1.0\n"
		"\n"
		"litefsd --pipe [ --fork ] [ --debug ] [ --chdir path ]\n"
		"litefsd --socket sock_path [--socket sock_path ] [ --fork ] [ --debug ] [ --chdir path ]\n"
		"litefsd --tcpport port [--tcpport port ] [ --fork ] [ --debug ] [ --chdir path ]\n");
}

static const char *workdir;
static int len_workdir;

struct ONEMOUNT{
	const char *path;
	int lenpath;
	bool is_mounted;
	time_t may_umount;	// At which time it may be umounted
};

class TBMOUNT{
public:
	bool silentmount;	// Use option -n
	ONEMOUNT tb[100];
	int nbmount;
	TBMOUNT(){
		nbmount = 0;
	}
	/*~PROTOBEG~ TBMOUNT */
public:
	void add (const char *path);
	void check_mount (const char *path);
	void check_umount (void);
	/*~PROTOEND~ TBMOUNT */
};

PUBLIC void TBMOUNT::add (const char *path)
{
	tb[nbmount].may_umount = 0;
	tb[nbmount].is_mounted = false;
	tb[nbmount].lenpath = strlen(path);
	tb[nbmount++].path = path;
}

PUBLIC void TBMOUNT::check_umount ()
{
	time_t t = time(NULL);
	ONEMOUNT *pt = tb;
	for (int i=0; i<nbmount; i++,pt++){
		if (pt->is_mounted && pt->may_umount <= t){
			char cmd[PATH_MAX];
			snprintf (cmd,sizeof(cmd)-1,"umount %s %s"
				,silentmount ? "-n" : ""
				,pt->path);
			n_debug ("executing %s\n",cmd);
			if (system(cmd)==0){
				pt->is_mounted = false;
			}else{
				pt->may_umount = t+3;
			}
		}
	}
}
PUBLIC void TBMOUNT::check_mount (const char *path)
{
	time_t t = time(NULL);
	ONEMOUNT *pt = tb;
	for (int i=0; i<nbmount; i++,pt++){
		int lenpath = pt->lenpath;
		if (strncmp(path,pt->path,lenpath)==0
			&& (path[lenpath] == '\0' || path[lenpath] == '/')){
			if (!pt->is_mounted){
				char cmd[PATH_MAX];
				snprintf (cmd,sizeof(cmd)-1,"mount %s %s"
					,silentmount ? "-n" : ""
					,pt->path);
				n_debug ("executing %s\n",cmd);
				system(cmd);
				pt->is_mounted = true;
			}
			pt->may_umount = t+2;
			break;
		}
	}
}


static void litefsd_2abs(
	const char *arg,
	char abspath[PATH_MAX])
{
	if (arg[0] == '/' && arg[1] == '/'){
		strncpy (abspath,arg,PATH_MAX);
	}else{
		strcpy (abspath,workdir);
		strncpy (abspath+len_workdir,arg,PATH_MAX-len_workdir);
	}
	abspath[PATH_MAX-1] = '\0';
}


/*
	Convert the request to an absolute path by matching a volume
	definition. Send failure to the client if there is no matching
	volume or the request is invalid (.. in path)

	Return -1 if this is not a valid request.
*/
static int litefsd_translate (
	int fd,
	LITE_REQUEST_DATA &req,
	char abspath[PATH_MAX],
	TBMOUNT &tbauto)
{
	int ret = -1;
	n_debug ("translate len = %d\n",req.len);
	if (req.len > 0){
		/* #Specification: litefs server / mapping the working dir
			litefs operates in the current directory unless it is forced
			to operate in the root. The users sees a litefs volume as
			a sub-directory (~/n/volume), so when it access a file
			~/n/volume/file is translate on the litefsd side to
			current_dir/file

			If the user specify a path like this: ~/n/volume//file, then
			this is converted to /file on the server side. This was done
			because it is expect that litefs will generally use to access
			a specific hierarchy on the server side, so by using
			the working directory of litefsd, we get this result.
		*/
		litefsd_2abs (req.buf,abspath);
		tbauto.check_mount(abspath);
		ret = 0;
	}else{
		litefsd_sendfail (fd);
	}
	return ret;
}

static void fct_child (int)
{
}

struct PROGCTL{
	int client;		// Handle to talk to the client
	int in,out,err;	// Handles to talk/listen to the program
					// as seen by the program
					// ("in" is used to write to the program standard input)
};

static int litefsd_setupexec (int fd, char *argv[], PROGCTL &ctl)
{
	int ret = -1;
	ctl.client = ctl.in = ctl.out = ctl.err = -1;
	signal (SIGCLD,fct_child);
	int tbin[2],tbout[2],tberr[2];
	if (pipe(tbin)!=-1
		&& pipe (tbout) != -1
		&& pipe (tberr) != -1){
		int pid = fork();
		if (pid == 0){
			dup2 (tbin[0],0);
			dup2 (tbout[1],1);
			dup2 (tberr[1],2);
			close (tbin[0]); close (tbin[1]);
			close (tbout[0]); close (tbout[1]);
			close (tberr[0]); close (tberr[1]);
			execv (argv[0],argv);
			_exit (-1);
		}else if (pid == -1){
		}else{
			close (tbin[0]);
			close (tbout[1]);
			close (tberr[1]);
			ctl.in = tbin[1];
			ctl.out = tbout[0];
			ctl.err = tberr[0];
			ctl.client = fd;
			ret = 0;
		}
	}
	return ret;
}

class  PROGCTLS{
	int first;
	int maxctls;
	PROGCTL *tb;
	/*~PROTOBEG~ PROGCTLS */
public:
	PROGCTLS (void);
	PROGCTL *alloc (int &id);
	void release (PROGCTL *ctl);
	/*~PROTOEND~ PROGCTLS */
};

PUBLIC PROGCTLS::PROGCTLS()
{
	first = 0;
	maxctls = 0;
	tb = NULL;
}

/*
	Return the first available PROGCTL
*/
PUBLIC PROGCTL *PROGCTLS::alloc (int &id)
{
	if (first == maxctls){
		int old = maxctls;
		maxctls += 10;
		tb = (PROGCTL *)realloc (tb,maxctls*sizeof(PROGCTL));
		PROGCTL *ctl = tb+old;
		for (int i=old; i<maxctls; i++, ctl++){
			ctl->client = i+1;
			ctl->in = ctl->out = ctl->err = -1;
		}
	}
	id = first;
	PROGCTL *ret = tb + first;
	first = ret->client;
	return ret;
}

PUBLIC void PROGCTLS::release (PROGCTL *ctl)
{
	ctl->client = first;
	ctl->in = ctl->out = ctl->err = -1;
	first = (int)(ctl - tb);
}

/*
	Read a request from one client
	Every request must contain a credential, to be added later
*/
static int litefsd_doreq (
	int fin,
	int fout,
	PROGCTLS &ctls,
	bool &redoset,		// Redo the set for select because ctls
						// have changes
	TBMOUNT &tbauto)
{
	redoset = false;
	int ret = -1;
	const int maxdir = 1024;
	static DIR *tbdir[maxdir];
	LITE_REQUEST_DATA req;
	int len = read (fin,&req,sizeof(req));
	if (len >= (int)sizeof(LITE_REQUEST)){
		ret = 0;
		len -= sizeof(LITE_REQUEST);
		while (len < req.len){
			int chunk = read (fin,req.buf+len,sizeof(req.buf)-len);
			if (chunk > 0){
				len += chunk;
			}else{
				ret = -1;
				break;
			}
		}
		if (req.cmd == LITE_VERSION){
			if (req.mode != LITEFS_PROTOCOL_VERSION){
				litefsd_sendfail (fout);
			}else{
				litefsd_sendok(fout);
			}
		}else if (req.cmd == LITE_SHUTDOWN){
			litefsd_sendok(fout);
			ret = -1;
		}else if (req.cmd == LITE_OPEN){
			n_debug ("open :%s: %d %d\n",req.buf,req.mode,req.perm);
			char abspath[PATH_MAX];
			if (litefsd_translate (fout,req,abspath,tbauto)!=-1){
				int handle = open (abspath,req.mode,req.perm);
				n_debug ("open abspath :%s: handle %d\n",abspath,handle);
				if (handle != -1){
					litefsd_sendfd (fout,handle);
				}else{
					litefsd_senderrno(fout);
				}
			}
		}else if (req.cmd == LITE_CLOSE){
			n_debug ("close %d\n",req.handle);
			if (close(req.handle)!=-1){
				litefsd_sendok(fout);
			}else{
				litefsd_senderrno(fout);
			}
		}else if (req.cmd == LITE_SEEK){
			n_debug ("seek %d %d %d\n",req.handle,req.mode,req.perm);
			off_t off = lseek(req.handle,req.mode,req.perm);
			if (off != (off_t)-1){
				litefsd_sendseek(fout,off);
			}else{
				litefsd_senderrno(fout);
			}
		}else if (req.cmd == LITE_READ){
			n_debug ("read %d %d\n",req.handle,req.mode);
			LITE_REPLY_DATA rep;
			rep.len = read (req.handle,rep.buf,req.mode);
			if (rep.len >= 0){
				rep.status = LITE_STATUS_OK;
				write (fout,&rep,sizeof(LITE_REPLY)+rep.len);
			}else{
				litefsd_senderrno(fout);
			}
		}else if (req.cmd == LITE_WRITE){
			int len = write (req.handle,req.buf,req.len);
			if (len >= 0){
				litefsd_sendok(fout,len);
			}else{
				litefsd_senderrno(fout);
			}
		}else if (req.cmd == LITE_STAT){
			n_debug ("stat :%s:\n",req.buf);
			char abspath[PATH_MAX];
			if (litefsd_translate (fout,req,abspath,tbauto)!=-1){
				n_debug ("stat abspath :%s:\n",abspath);
				struct stat st;
				if (stat(abspath,&st)==-1){
					litefsd_senderrno(fout);
				}else{
					litefsd_sendstat (fout,st);
				}
			}
		}else if (req.cmd == LITE_LSTAT){
			n_debug ("lstat :%s:\n",req.buf);
			char abspath[PATH_MAX];
			if (litefsd_translate (fout,req,abspath,tbauto)!=-1){
				n_debug ("lstat abspath :%s:\n",abspath);
				struct stat st;
				if (lstat(abspath,&st)==-1){
					litefsd_senderrno(fout);
				}else{
					litefsd_sendstat (fout,st);
				}
			}
		}else if (req.cmd == LITE_FSTAT){
			struct stat st;
			if (fstat(req.handle,&st)==-1){
				n_debug ("fstat %d failed %d\n",req.handle,errno);
				litefsd_senderrno(fout);
			}else{
				litefsd_sendstat (fout,st);
				n_debug ("fstat %d ok\n",req.handle);
			}
		}else if (req.cmd == LITE_UNLINK){
			n_debug ("unlink :%s:\n",req.buf);
			char abspath[PATH_MAX];
			if (litefsd_translate (fout,req,abspath,tbauto)!=-1){
				n_debug ("unlink abspath :%s:\n",abspath);
				if (unlink(abspath)!=-1){
					litefsd_sendok (fout);
				}else{
					litefsd_senderrno (fout);
				}
			}
		}else if (req.cmd == LITE_RENAME){
			const char *newp = req.buf;
			while (*newp != '\0') newp++;
			newp++;
			n_debug ("rename :%s: :%s:\n",req.buf,newp);
			char old_abspath[PATH_MAX];
			if (litefsd_translate (fout,req,old_abspath,tbauto)!=-1){
				char new_abspath[PATH_MAX];
				litefsd_2abs (newp,new_abspath);
				n_debug ("rename abspath :%s: :%s:\n",old_abspath,new_abspath);
				if (rename(old_abspath,new_abspath)!=-1){
					litefsd_sendok (fout);
				}else{
					litefsd_senderrno (fout);
				}
			}
		}else if (req.cmd == LITE_LINK){
			const char *newp = req.buf;
			while (*newp != '\0') newp++;
			newp++;
			n_debug ("link :%s: :%s:\n",req.buf,newp);
			char old_abspath[PATH_MAX];
			if (litefsd_translate (fout,req,old_abspath,tbauto)!=-1){
				char new_abspath[PATH_MAX];
				litefsd_2abs (newp,new_abspath);
				n_debug ("link abspath :%s: :%s:\n",old_abspath,new_abspath);
				if (link(old_abspath,new_abspath)!=-1){
					litefsd_sendok (fout);
				}else{
					litefsd_senderrno (fout);
				}
			}
		}else if (req.cmd == LITE_SYMLINK){
			const char *oldp = req.buf;
			while (*oldp != '\0') oldp++;
			oldp++;
			n_debug ("link :%s: :%s:\n",oldp,req.buf);
			char new_abspath[PATH_MAX];
			if (litefsd_translate (fout,req,new_abspath,tbauto)!=-1){
				n_debug ("symlink abspath :%s: :%s:\n",oldp,new_abspath);
				if (symlink(oldp,new_abspath)!=-1){
					litefsd_sendok (fout);
				}else{
					litefsd_senderrno (fout);
				}
			}
		}else if (req.cmd == LITE_UTIMES){
			n_debug ("utimes :%s:\n",req.buf);
			char abspath[PATH_MAX];
			if (litefsd_translate (fout,req,abspath,tbauto)!=-1){
				n_debug ("utimes abspath :%s: %d %d\n",abspath,req.mode,req.perm);
				int ok = -1;
				if (req.mode == -1 && req.perm == -1){
					ok = utimes (abspath,NULL);
				}else{
					struct timeval tvp;
					tvp.tv_sec = req.mode;
					tvp.tv_usec = req.perm;
					ok = utimes (abspath,&tvp);
				}
				if (ok!=-1){
					litefsd_sendok (fout);
				}else{
					litefsd_senderrno (fout);
				}
			}
		}else if (req.cmd == LITE_ACCESS){
			n_debug ("access :%s:\n",req.buf);
			char abspath[PATH_MAX];
			if (litefsd_translate (fout,req,abspath,tbauto)!=-1){
				n_debug ("access abspath :%s: %d %d\n",abspath,req.mode,req.perm);
				int ok = access (abspath,req.mode);
				if (ok!=-1){
					litefsd_sendok (fout);
				}else{
					litefsd_senderrno (fout);
				}
			}
		}else if (req.cmd == LITE_RMDIR){
			n_debug ("rmdir :%s:\n",req.buf);
			char abspath[PATH_MAX];
			if (litefsd_translate (fout,req,abspath,tbauto)!=-1){
				n_debug ("rmdir abspath :%s:\n",abspath);
				if (rmdir(abspath)!=-1){
					litefsd_sendok (fout);
				}else{
					litefsd_senderrno (fout);
				}
			}
		}else if (req.cmd == LITE_MKDIR){
			n_debug ("mkdir :%s:\n",req.buf);
			char abspath[PATH_MAX];
			if (litefsd_translate (fout,req,abspath,tbauto)!=-1){
				n_debug ("mkdir abspath :%s:\n",abspath);
				if (mkdir(abspath,req.mode)!=-1){
					litefsd_sendok (fout);
				}else{
					litefsd_senderrno (fout);
				}
			}
		}else if (req.cmd == LITE_CHDIR){
			// chdir means, may chdir
			char abspath[PATH_MAX];
			if (litefsd_translate (fout,req,abspath,tbauto)!=-1){
				n_debug ("chdir abspath :%s:\n",abspath);
				struct stat st;
				if (stat(abspath,&st)!=-1 && S_ISDIR(st.st_mode)){
					litefsd_sendok (fout);
				}else{
					litefsd_senderrno (fout);
				}
			}
		}else if (req.cmd == LITE_CHMOD){
			char abspath[PATH_MAX];
			if (litefsd_translate (fout,req,abspath,tbauto)!=-1){
				n_debug ("chmod abspath :%s: %o\n",abspath,req.mode);
				if (chmod (abspath,req.mode)!=-1){
					litefsd_sendok (fout);
				}else{
					litefsd_senderrno (fout);
				}
			}
		}else if (req.cmd == LITE_FCHMOD){
			n_debug ("fchmod %d\n",req.handle);
			if (fchmod(req.handle,req.mode)!=-1){
				litefsd_sendok (fout);
			}else{
				litefsd_senderrno (fout);
			}
		}else if (req.cmd == LITE_FCHOWN){
			n_debug ("fchown %d\n",req.handle);
			if (fchown(req.handle,req.mode,req.perm)!=-1){
				litefsd_sendok (fout);
			}else{
				litefsd_senderrno (fout);
			}
		}else if (req.cmd == LITE_CHOWN){
			char abspath[PATH_MAX];
			if (litefsd_translate (fout,req,abspath,tbauto)!=-1){
				n_debug ("chown abspath :%s: %d %d\n",abspath,req.mode,req.perm);
				if (chown (abspath,req.mode,req.perm)!=-1){
					litefsd_sendok (fout);
				}else{
					litefsd_senderrno (fout);
				}
			}
		}else if (req.cmd == LITE_OPENDIR){
			n_debug ("opendir :%s:\n",req.buf);
			char abspath[PATH_MAX];
			if (litefsd_translate (fout,req,abspath,tbauto)!=-1){
				n_debug ("opendir abspath :%s:\n",abspath);
				DIR *dir = opendir (abspath);
				if (dir != NULL){
					int i;
					for (i=0; i<maxdir; i++){
						if (tbdir[i] == NULL){
							tbdir[i] = dir;
							litefsd_sendhandle (fout,i);
							break;
						}
					}
					if (i==maxdir) litefsd_sendfail (fout);
				}else{
					litefsd_senderrno (fout);
				}
			}
		}else if (req.cmd == LITE_READDIR){
			n_debug ("readdir %d\n",req.handle);
			DIR *dir = tbdir[req.handle];
			if (dir == NULL){
				litefsd_sendfail (fout);
			}else{
				struct dirent *ent = readdir (dir);
				LITE_REPLY_DATA rep;
				rep.status = LITE_STATUS_OK;
				if (ent == NULL){
					rep.len = 0;
				}else{
					rep.len = sizeof(struct dirent);
					memcpy (rep.buf,ent,sizeof(struct dirent));
				}
				write (fout,&rep,sizeof(LITE_REPLY)+rep.len);
			}
		}else if (req.cmd == LITE_CLOSEDIR){
			n_debug ("closedir %d\n",req.handle);
			DIR *dir = tbdir[req.handle];
			if (dir == NULL){
				litefsd_sendfail (fout);
			}else if (closedir(dir)!=-1){
				litefsd_sendok(fout);
			}else{
				litefsd_senderrno(fout);
			}
		}else if (req.cmd == LITE_READLINK){
			n_debug ("readlink %d\n",req.handle);
			char abspath[PATH_MAX];
			if (litefsd_translate (fout,req,abspath,tbauto)!=-1){
				n_debug ("readlink abspath :%s:\n",abspath);
				LITE_REPLY_DATA rep;
				rep.len = readlink (abspath,rep.buf,sizeof(rep.buf));
				if (len ==-1){
					litefsd_senderrno (fout);
				}else{
					rep.status = LITE_STATUS_OK;
					write (fout,&rep,sizeof(LITE_REPLY)+rep.len);
				}
			}
		}else if (req.cmd == LITE_EXEC){
			n_debug ("exec :%s:\n",req.buf);
			char *pt = req.buf;
			while (*pt > ' ') pt++;
			char abspath[PATH_MAX];
			char *argv[1000];
			argv[0] = abspath;
			int nb = 1;
			if(isspace(*pt)){
				*pt++ = '\0';
				while (1){
					while (isspace(*pt)) pt++;
					if (*pt == '\0') break;
					argv[nb] = pt;
					while (*pt > ' ') pt++;
					if (isspace(*pt)) *pt++ = '\0';
				}
			}
			argv[nb] = NULL;
			if (litefsd_translate (fout,req,abspath,tbauto)!=-1){
				int id;
				PROGCTL *ctl = ctls.alloc(id);
				if (litefsd_setupexec (fout,argv,*ctl) != -1){
					LITE_REPLY rep;
					rep.status = LITE_STATUS_OK;
					rep.data = id;
					rep.len = 0;
					write (fout,&rep,sizeof(LITE_REPLY));
					redoset = true;
				}else{
					ctls.release (ctl);
					litefsd_sendfail (fout);
				}
			}else{
				litefsd_sendfail (fout);
			}
		}else{
			litefsd_sendfail(fout);
			ret = -1;
		}
	}
	return ret;
}
		
struct ONESOCK{
	int handle;
	int family;
};

class TBSOCK{
public:
	ONESOCK tb[100];
	int nbsock;
	TBSOCK(){
		nbsock = 0;
	}
	/*~PROTOBEG~ TBSOCK */
public:
	void add (int family, int sock);
	int fdset (fd_set&fdin, int maxfd);
	/*~PROTOEND~ TBSOCK */
};

PUBLIC void TBSOCK::add (int family, int sock)
{
	tb[nbsock].family = family;
	tb[nbsock++].handle = sock;
}

/*
	Fill an fd_set with all the listening socket handle.
	Return the new maxfd
*/
PUBLIC int TBSOCK::fdset (fd_set &fdin, int maxfd)
{
	for (int i=0; i<nbsock; i++){
		int sock = tb[i].handle;
		FD_SET (sock,&fdin);
		if (sock >= maxfd) maxfd = sock + 1;
	}
	return maxfd;
}


/*
	Process all request and connections. Infinite loop
*/
static void litefsd_loop (
	TBSOCK &tbsock,		// Unix/Tcp socket to receive new connection
	TBMOUNT &tbauto,	// Directory to mount/umount as needed
	int pipein,			// Pre-connected client or proxy
	int pipeout)		// Pre-connected client or proxy
{
	int tbfd[1024];
	int nbfd = 0;
	fd_set fdin;
	FD_ZERO (&fdin);
	int maxfd = tbsock.fdset (fdin,0);
	if (pipein != -1){
		FD_SET (pipein,&fdin);
		if (pipein >= maxfd) maxfd = pipein+1;
	}
	PROGCTLS ctls;
	while (1){
		fd_set in;
		memcpy (&in,&fdin,sizeof(in));
		n_debug ("select maxfd %d\n",maxfd);
		struct timeval timeout;
		timeout.tv_sec = 2;
		timeout.tv_usec = 0;
		if (select (maxfd,&in,NULL,NULL,&timeout)!=-1){
			int orig_nbfd = nbfd;
			bool redo_fdin = false;
			for (int i=0; i<tbsock.nbsock; i++){
				int sock = tbsock.tb[i].handle;
				if (FD_ISSET (sock,&in)){
					struct sockaddr_un unc;
					size_t len = sizeof(unc);
					unc.sun_family = tbsock.tb[i].family;
					int s = accept (sock,(struct sockaddr*)&unc,&len);
					tbfd[nbfd++] = s;
					FD_SET (s,&fdin);
					if (s >= maxfd) maxfd = s + 1;
				}
			}
			if (pipein != -1 && FD_ISSET(pipein,&in)){
				if (litefsd_doreq (pipein,pipeout,ctls,redo_fdin,tbauto)
					== -1){
					break;
				}
			}
			int *ptfd = tbfd;
			for (int i=0; i< orig_nbfd; i++, ptfd++){
				int fd = *ptfd;
				if (FD_ISSET(fd,&in)){
					if (litefsd_doreq (fd,fd,ctls,redo_fdin,tbauto) == -1){
						close (fd);
						redo_fdin = true;
						*ptfd = -1;
					}
				}
			}
			if (redo_fdin){
				FD_ZERO (&fdin);
				maxfd = tbsock.fdset (fdin,0);
				if (pipein != -1){
					FD_SET (pipein,&fdin);
					if (pipein >= maxfd) maxfd = pipein+1;
				}
				ptfd = tbfd;
				int *copyfd = tbfd;
				for (int i=0; i<nbfd; i++, ptfd++){
					int fd = *ptfd;
					if (fd != -1){
						FD_SET (fd,&fdin);
						if (fd >= maxfd) maxfd = fd + 1;
						*copyfd++ = fd;
					}
				}
				nbfd = (int)(copyfd - tbfd);
			}
		}
		tbauto.check_umount();
	}
}

int main (int argc, char *argv[])
{
	int ret = -1;
	if (argc == 1){
		usage();
	}else{
		TBSOCK tbsock;		// Listening sockets
		TBMOUNT tbauto;		// Dir that should be automounted
		int pipein = -1;
		int pipeout = -1;
		const char *newdir = NULL;
		bool dofork = false;
		for (int i=1; i<argc; i++){
			if (strcmp(argv[i],"--socket")==0){
				i++;
				int sock = litefsd_server(argv[i]);
				if (sock != -1){
					tbsock.add (AF_UNIX,sock);
				}
			}else if (strcmp(argv[i],"--pipe")==0){
				pipein = 0;
				pipeout = 1;
			}else if (strcmp(argv[i],"--tcpport")==0){
				i++;
				int sock = litefsd_tcpserver(argv[i]);
				if (sock != -1){
					tbsock.add (AF_INET,sock);
				}
			}else if (strcmp(argv[i],"--automount")==0){
				i++;
				tbauto.add (argv[i]);
			}else if (strcmp(argv[i],"--chdir")==0){
				i++;
				newdir = argv[i];
			}else if (strcmp(argv[i],"--debug")==0){
				debug = true;
			}else if (strcmp(argv[i],"--fork")==0){
				dofork = true;
			}else{
				usage();
				exit (-1);
			}
		}
		if (newdir != NULL && chdir(newdir)==-1){
			fprintf (stderr,"Can't change to directory %s (%s)\n"
				,newdir,strerror(errno));
			exit (-1);
		}
		char cwd[PATH_MAX];
		if (getcwd(cwd,PATH_MAX)==NULL){
			fprintf (stderr,"Can't locate current working directory (%s)\n"
				,strerror(errno));
		}else{
			workdir = cwd;
			len_workdir = strlen(cwd);
			if (tbsock.nbsock > 0 || pipein != -1){
				if (dofork){
					if (fork() != 0) exit(0);
					setsid();
				}
				litefsd_loop(tbsock,tbauto,pipein,pipeout);
			}else{
				usage();
			}
		}
	}
	return ret;
}



